import numpy as np
from qtpyt.lo.tools import extract_orthogonal_subspaces

def test_U1():
    h = np.arange(16).reshape(4,4)
    s = np.eye(4)
    bfs_m = [1]

    hs_mm, hs_ii, hs_im, U = extract_orthogonal_subspaces(h, s, bfs_m)
    np.testing.assert_allclose(hs_mm[0], h[1,1])
    np.testing.assert_allclose(hs_ii[0], h[np.ix_([0,2,3],[0,2,3])])
    np.testing.assert_allclose(hs_im[0], h[np.ix_([0,2,3],[1])])

def test_U2():
    h = np.random.random((4,4))
    s = np.random.random((4,4))
    s += s.T
    s /= 2
    bfs_m = [1]

    hs_mm, hs_ii, hs_im, U = extract_orthogonal_subspaces(h, s, bfs_m)
    np.testing.assert_allclose(hs_mm[0], h[1,1])
    np.testing.assert_allclose(hs_im[1], np.zeros((4-1,1)), atol=1e-16)


"""Simple tight-binding add-on for GPAWs LCAO module."""

from math import pi

import numpy as np
import scipy.linalg as sla

from qtpyt.surface.kpts import (apply_inversion_symm, 
                    expand_inversion_symm, fourier_sum, 
                    get_monkhorst_pack_size_and_offset)


def bloch_to_realspace(A, kpts, R, out=None):
    '''K-space Fourier sum.
    
    The main difference w.r.t. fourier_sum from qtpyt.surface.kpts
    is that it accetps A's k-matrices which are evaluated in the irreducible
    BZ. In this case, kpts MUST be a reducible set of kpts.

    Args:
        A : (np.ndarray) matrices in k-space.
        kpts : (1D or 2D) k-points at which A's have been evaluated.
        R : (1D or 2D) lattice vectors at which A's will be transformed.
    '''
    size, offset = get_monkhorst_pack_size_and_offset(kpts)
    if not np.allclose(offset, (0,0,0)):
        # TODO :: In this case one cannot simply out.conj()
        raise NotImplementedError

    kpts = np.atleast_2d(kpts)
    R = np.atleast_2d(R)
    # Irreducible brillouin zone
    ibz = False
    if kpts.shape[0] > A.shape[0]:
        # A's are at IBZ
        Nk = [len(set(kpts[:,i])) for i in range(3)]
        kpts = apply_inversion_symm(kpts)[0]
        ibz = True
    out = fourier_sum(A, kpts, R, out)
    if ibz:
        out *= kpts.shape[0]
        out += out.conj()
        try:
            gamma = np.where(np.sum(kpts, axis=1)<1e-9)[0][0]
        except IndexError:
            pass
        else:
            #       ik          -ik 
            # A(k) e    + A(-k) e   
            #                                 T    *
            # NOTE: A(k) = A(-k)  & A(k) = ( A(k) )
            #
            #         ik   *     *   -ik       T    * -ik          -ik
            # ( A(k) e    )  =  A(k) e   = ( A(-k) )  e    = A(-k) e  
            out -= A[gamma]
        out /= np.product(Nk)
    return out


class TightBinding:
    """Simple class for tight-binding calculations.
    
    Example:
        kpts = monkhorst-pack(size)
        h_k :: Hamilton evaulated at IBZ
        s_k :: Overlap evaulated at IBZ
        tb = TightBinding(kpts, h_k, s_k)
    """

    def __init__(self, kpts, H_kMM, S_kMM):
        """
        Args:
            kpts : (1D or 2D) 
                k-points at which A's have been evaluated.
            H_kMM, S_kMM : (np.ndarray) 
                Hamiltonian and overlap matrices in k-space.
        """
        if H_kMM.shape[0] < kpts.shape[0]:
            ks, H_kMM = expand_inversion_symm(kpts, H_kMM)
            assert np.allclose(ks, kpts), 'Invalid k-points'
            S_kMM = expand_inversion_symm(kpts, S_kMM)[1]

        self.kpts = kpts
        # here we have set kpts to the reducible representation,
        # so Nk is correctly computed when called from set_num_cells.
        self.set_num_cells()
        self.H_NMM = fourier_sum(H_kMM, self.kpts, self.R_cN.T)
        self.S_NMM = fourier_sum(S_kMM, self.kpts, self.R_cN.T)

    @property
    def Nk(self):
        """# of transverse k-points"""
        return np.array([len(set(self.kpts[:,i])) for i in range(3)])

    def set_num_cells(self, N_c=None):
        """Set number of real-space cells to use.

        Parameters
        ----------
        N_c: tuple or ndarray
            Number of unit cells in each direction of the basis vectors.

        """

        if N_c is None:
            self.N_c = tuple(self.Nk)
        else:
            self.N_c = tuple(N_c)
        try:
            if np.any(np.asarray(self.Nk) < np.asarray(self.N_c)):
                print("WARNING: insufficient k-point sampling.")
        except AttributeError:
            # Nk is not defined
            pass
        # Lattice vectors
        R_cN = np.indices(self.N_c).reshape(3, -1)
        N_c = np.array(self.N_c)[:, np.newaxis]
        R_cN += N_c // 2
        R_cN %= N_c
        R_cN -= N_c // 2
        self.R_cN = R_cN

    def band_structure(self, path_kc, blochstates=False):
        """Calculate dispersion along a path in the Brillouin zone.

        Parameters
        ----------
        path_kc: ndarray
            List of k-point coordinates (in units of the reciprocal lattice
            vectors) specifying the path in the Brillouin zone for which the
            dynamical matrix will be calculated.
        blochstates: bool
            Return LCAO expansion coefficients when True (default: False).

        """
        R_cN = self.R_cN
        # Lists for eigenvalues and eigenvectors along path
        eps_kn = []
        psi_kn = []

        for k_c in path_kc:
            # Evaluate fourier sum
            phase_N = np.exp(-2.j * pi * np.dot(k_c, R_cN))
            H_MM = np.sum(phase_N[:, np.newaxis, np.newaxis] * self.H_NMM,
                          axis=0)
            S_MM = np.sum(phase_N[:, np.newaxis, np.newaxis] * self.S_NMM,
                          axis=0)

            if blochstates:
                eps_n, c_Mn = sla.eigh(H_MM, S_MM)
                # Sort eigenmodes according to increasing eigenvalues
                c_nM = c_Mn[:, eps_n.argsort()].transpose()
                psi_kn.append(c_nM)
            else:
                eps_n = sla.eigvalsh(H_MM, S_MM)

            # Sort eigenvalues in increasing order
            eps_n.sort()
            eps_kn.append(eps_n)

        # Convert to eV
        eps_kn = np.array(eps_kn) #* units.Hartree

        if blochstates:
            return eps_kn, np.array(psi_kn)

        return eps_kn



# cython: infer_types=True
# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False

import numpy as np
from scipy import linalg as la
from cython.parallel import prange, parallel
from scipy.linalg.cython_lapack cimport zgetrf, zgetrs, zgetri, zlange, zlacpy
from scipy.linalg.cython_blas cimport zgemm, zaxpy


cdef double conv = 1e-8


cdef inline complex conj(complex a) nogil:
    return a.real - a.imag*1j


cdef inline void init_vecs(complex* h_ii, complex* s_ii,
                           complex[:,:] h_ij, complex[:,:] s_ij,
                           complex* v_00, complex* v_01,
                           complex* v_10, complex* v_11,
                           complex z, int m) nogil:

    cdef int i, j

    for i in prange(m, schedule='static'):
        for j in range(m):
            v_00[i*m+j] = z * conj(s_ii[i*m+j]) - conj(h_ii[i*m+j])
            v_11[i*m+j] = v_00[i*m+j]
            v_01[i*m+j] = z * conj(s_ij[i,j]) - conj(h_ij[i,j])

    for i in prange(m, schedule='static'):
        for j in range(m):
            v_10[i*m+j] = z * s_ij[j,i] - h_ij[j,i]


cdef void get_Ginv(complex* v_00, complex* v_01,
                   complex* v_10, complex* v_11,
                   complex[::1,:] a, complex[::1,:] b,
                   complex[::1,:] lu, complex[::1,:] v_01_dot_b,
                   int[:] ipiv, int m) nogil:
    # """The inverse of the retarded surface Green function"""

    cdef int i, j, info, inc = 1
    cdef int m2 = m*m
    cdef char trans = b'N', norm = b'M', UPLO = b'A'
    cdef complex alpha = 1.+0.j, beta = -1.+0.j, gamma = 0.+0.j

    cdef double delta = conv + 1
    cdef double work[0]

    while delta > conv:

        zlacpy(&UPLO,&m,&m,&v_11[0],&m,&lu[0,0],&m)
        zgetrf(&m,&m,&lu[0,0],&m,&ipiv[0],&info)
        zlacpy(&UPLO,&m,&m,&v_01[0],&m,&a[0,0],&m)
        zgetrs(&trans,&m,&m,&lu[0,0],&m,&ipiv[0],&a[0,0],&m,&info);
        zlacpy(&UPLO,&m,&m,&v_10[0],&m,&b[0,0],&m)
        zgetrs(&trans,&m,&m,&lu[0,0],&m,&ipiv[0],&b[0,0],&m,&info);
        zgemm(&trans,&trans,&m,&m,&m,&alpha,&v_01[0],&m,&b[0,0],
              &m,&gamma,&v_01_dot_b[0,0],&m)
        zaxpy(&m2,&beta,&v_01_dot_b[0,0],&inc,&v_00[0],&inc)
        zgemm(&trans,&trans,&m,&m,&m,&beta,&v_10[0],&m,&a[0,0],
              &m,&alpha,&v_11[0],&m)
        zaxpy(&m2,&beta,&v_01_dot_b[0,0],&inc,&v_11[0],&inc)
        zlacpy(&UPLO,&m,&m,&v_01[0],&m,&lu[0,0],&m)
        zgemm(&trans,&trans,&m,&m,&m,&beta,&lu[0,0],&m,&a[0,0],
              &m,&gamma,&v_01[0],&m)
        zlacpy(&UPLO,&m,&m,&v_10[0],&m,&lu[0,0],&m)
        zgemm(&trans,&trans,&m,&m,&m,&beta,&lu[0,0],&m,&b[0,0],
              &m,&gamma,&v_10[0],&m)
        delta = zlange(&norm,&m,&m,&v_01[0],&m,&work[0])

cdef inline void inv(complex* a, complex[::1] work, int[:] ipiv, int m) nogil:

    cdef int info, lwork = work.shape[0]

    zgetrf(&m,&m,&a[0],&m,&ipiv[0],&info)
    zgetri(&m,&a[0],&m,&ipiv[0],&work[0],&lwork,&info)


def get_G(G_kMM, H_kii, S_kii, H_kij, S_kij, double energy, 
          double eta=1e-5, double bias=0.):

    nkpts = H_kii.shape[0]
    m = G_kMM.shape[1]
    n = G_kMM.shape[2]
    lwork = 32000

    # v_xx = z * OP(h_xx) - OP(s_xx)
    # OP :: .T or .H
    v_00 = np.zeros((m,n), complex, order='F')
    cdef complex[:,:] _v_00 = v_00
    v_01 = np.zeros((m,n), complex, order='F')
    cdef complex[:,:] _v_01 = v_01
    v_10 = np.zeros((m,n), complex, order='F')
    cdef complex[:,:] _v_10 = v_10
    v_11 = np.zeros((m,n), complex, order='F')
    cdef complex[:,:] _v_11 = v_11

    # Loop internal vars.
    a = np.zeros((m,n), complex, order='F')
    b = np.zeros((m,n), complex, order='F')
    lu = np.zeros((m,n), complex, order='F')
    v_01_dot_b = np.zeros((m,n), complex, order='F')

    # Lapack internal vars.
    ipiv = np.zeros(m, np.int32)
    work = np.zeros(lwork, complex)

    # Green's function at transverse k-points
    cdef complex[:,:] h_ii
    cdef complex[:,:] s_ii
    cdef complex[:,:] h_ij
    cdef complex[:,:] s_ij
    cdef complex z
    z = energy - bias + eta * 1.j
    for i in range(nkpts):
        h_ii = H_kii[i]
        s_ii = S_kii[i]
        h_ij = H_kij[i]
        s_ij = S_kij[i]
        init_vecs(&h_ii[0,0],&s_ii[0,0],H_kij[i],S_kij[i],
                  &_v_00[0,0],&_v_01[0,0],&_v_10[0,0],&_v_11[0,0],
                  z,m)
        get_Ginv(&_v_00[0,0],&_v_01[0,0],&_v_10[0,0],&_v_11[0,0],
                 a,b,lu,v_01_dot_b,ipiv,m)
        inv(&_v_00[0,0],work,ipiv,m)
        G_kMM[i] = v_00



import numpy as np
from math import sqrt
from scipy import linalg as la


dagger = lambda mat: np.conj(mat.T)


def rotate_matrix(h, u):
    return np.dot(u.T.conj(), np.dot(h, u))


def get_subspace(matrix, index):
    """Get the subspace spanned by the basis function listed in index"""
    assert matrix.ndim == 2 and matrix.shape[0] == matrix.shape[1]
    return matrix.take(index, 0).take(index, 1)


def normalize(matrix, S=None):
    """Normalize column vectors.

    ::

      <matrix[:,i]| S |matrix[:,i]> = 1

    """
    for col in matrix.T:
        if S is None:
            col /= np.linalg.norm(col)
        else:
            col /= np.sqrt(np.dot(col.conj(), np.dot(S, col)))


def get_orthonormal_subspace(h_ii, s_ii, index_j=None):
    """Get orthonormal eigen-values and -vectors of subspace j."""
    if index_j is not None:
        h_sub_jj  = get_subspace(h_ii, index_j)
        s_sub_jj  = get_subspace(s_ii, index_j)
    else:
        h_sub_jj = h_ii
        s_sub_jj = s_ii
    e_j, v_jj = np.linalg.eig(np.linalg.solve(s_sub_jj, h_sub_jj))
    normalize(v_jj, s_sub_jj)  # normalize: <v_j|s|v_j> = 1
    permute_list = np.argsort(e_j.real)
    e_j = np.take(e_j, permute_list)
    v_jj = np.take(v_jj, permute_list, axis=1)
    return v_jj, e_j


def subdiagonalize(h_ii, s_ii, index_j):
    """Subdiagonalize a block in the Hamiltonian."""
    nb = h_ii.shape[0]
    nb_sub = len(index_j)
    v_jj, e_j = get_orthonormal_subspace(h_ii, s_ii, index_j)
    # Setup transformation matrix
    c_ii = np.identity(nb, h_ii.dtype)
    for i in range(nb_sub):
        for j in range(nb_sub):
            c_ii[index_j[i], index_j[j]] = v_jj[i, j]

    h1_ii = rotate_matrix(h_ii, c_ii)
    s1_ii = rotate_matrix(s_ii, c_ii)

    return h1_ii, s1_ii, c_ii, e_j


def subdiagonalize_atoms(basis, h_ii, s_ii, a=None):
    """Get rotation matrix for subdiagonalization of given(all) atoms."""
    from scipy.linalg import block_diag
    if a is None:
        a = range(len(basis.atoms))
    if isinstance(a, int):
        a = [a]
    U_mm = []
    e_aj = []
    na_tot = len(basis.atoms)
    for a0 in range(na_tot):
        bfs = basis.get_indices(a0)
        if a0 in a:
            v_jj, e_j = get_orthonormal_subspace(h_ii,s_ii,bfs)
            U_mm.append(v_jj)
            e_aj.append(e_j)
        else:
            U_mm.append(np.eye(len(bfs)))
    U_mm = block_diag(*U_mm)
    return U_mm, e_aj


def cutcoupling(h, s, index_n):
    for i in index_n:
        s[:, i] = 0.0
        s[i, :] = 0.0
        s[i, i] = 1.0
        Ei = h[i, i]
        h[:, i] = 0.0
        h[i, :] = 0.0
        h[i, i] = Ei


def lowdin_rotation(h, s, index_j=None):
    if index_j:
        s_jj = get_subspace(s, index_j)
    else:
        s_jj = s
    eig, U_jj = np.linalg.eigh(s_jj)
    eig = np.abs(eig)
    U_jj = np.dot(U_jj / np.sqrt(eig), dagger(U_jj))
    return U_jj


def get_U1(bfs_m, bfs_i, c_MM=None):
    """Get rotation matrix to extract sub-block defined 
    by matrix indices `bfs_m` to the bottom rigth of the 
    matrix. If a rotation c_MM is given apply permutation to c_MM.

    """
    nbf = len(bfs_m) + len(bfs_i)
    if c_MM is None:
        c_MM = np.eye(nbf)
    U1 = np.take(c_MM, np.concatenate([bfs_i, bfs_m]), axis=1)
    """If apply is True permute inplace bfs_m bfs_i"""
    bfs_m[:] = range(nbf-len(bfs_m), nbf)
    bfs_i[:] = range(nbf-len(bfs_m))
    return U1, bfs_m, bfs_i


def get_U2(s1_MM, bfs_m, bfs_i):
    """Get rotation matrix that orthogonalizes sub-blocks spanned
    by indices `bfs_m` and `bfs_i`, respectively.
    
    """
    nbf = len(bfs_m) + len(bfs_i)
    s_mm = get_subspace(s1_MM, bfs_m)
    s_mi = s1_MM.take(bfs_m,axis=0).take(bfs_i,axis=1)
    U_mi = - np.linalg.inv(s_mm).dot(s_mi)
    U2 = np.eye(s1_MM.shape[0])
    U2[np.ix_(bfs_m,bfs_i)] = U_mi
    return U2


def get_U2T(s1_MM, bfs_m, bfs_i):
    """Get rotation matrix that orthogonalizes sub-blocks spanned
    by indices `bfs_m` and `bfs_i`, respectively.
    
    """
    nbf = len(bfs_m) + len(bfs_i)
    s_ii = get_subspace(s1_MM, bfs_i)
    s_im = s1_MM.take(bfs_i,axis=0).take(bfs_m,axis=1)
    U_im = - np.linalg.inv(s_ii).dot(s_im)
    U2T = np.eye(s1_MM.shape[0])
    U2T[np.ix_(bfs_i,bfs_m)] = U_im
    return U2T


def extract_orthogonal_subspaces(h_MM, s_MM, bfs_m, c_MM=None, T=False):
    nbf = h_MM.shape[0]
    bfs_i = np.setdiff1d(range(nbf), bfs_m)
    U1, bfs_m, bfs_i = get_U1(bfs_m, bfs_i, c_MM)
    h1_MM = rotate_matrix(h_MM, U1)
    s1_MM = rotate_matrix(s_MM, U1)
    U2 = get_U2(s1_MM, bfs_m, bfs_i) if not T else get_U2T(s1_MM, bfs_m, bfs_i)
    h2_MM = rotate_matrix(h1_MM, U2)
    s2_MM = rotate_matrix(s1_MM, U2)
    hs_mm, hs_ii, hs_im = split_subspaces(h2_MM, s2_MM, bfs_m, bfs_i)
    U = U1.dot(U2)
    return hs_mm, hs_ii, hs_im, U

                 
def split_subspaces(h_MM, s_MM, bfs_m, bfs_i):
    nbf = h_MM.shape[0]
    h_ii = get_subspace(h_MM, bfs_i)
    s_ii = get_subspace(s_MM, bfs_i)
    h_mm = get_subspace(h_MM, bfs_m)
    s_mm = get_subspace(s_MM, bfs_m)
    h_im = h_MM.take(bfs_i,axis=0).take(bfs_m,axis=1)
    s_im = s_MM.take(bfs_i,axis=0).take(bfs_m,axis=1)
    return (h_mm, s_mm), (h_ii, s_ii), (h_im, s_im)


def flatten(iterables):
    return (elem for iterable in iterables for elem in iterable)


def get_orthogonal_subspaces(basis, h_MM, s_MM, a=None, cutoff=np.inf):
    """Split matrix in hybridization and active spaces.
       Active space contains orbital(s) of atom(s) a with
       abs(energy) < cutoff"""
    nbf = len(h_MM)
    c_MM, e_aj = subdiagonalize_atoms(basis, h_MM, s_MM, a)
    bfs_imp = basis.get_indices(a)
    bfs_m = [bfs_imp[i] for i, eigval in enumerate(flatten(e_aj)) if abs(eigval)<cutoff]
    return extract_orthogonal_subspaces(h_MM, s_MM, bfs_m, c_MM)


def order_transform(basis, groups=None, nuc=None):
    """Order orbital indices using group of atoms and atomic elements.
    
    For each subset of atoms, group the atoms of the same kind and order
    the orbital indices in scending order. When both `groups` and `nuc` 
    are None, treat the whole set of atoms as a single group.

    Args:
        basis : (Basis object)
        groups : (list of lists) 
            List of the atoms in each group.
        nuc : (int)
            Number of sequential unit cells in atoms.
            Each unit cell is considered a group.
    """
    atoms = basis.atoms
    symbols = atoms.symbols

    elements = set(symbols)
    natoms = len(atoms)
    
    if groups is not None:
        try:
            groups[0][0]
        except IndexError:
            raise ValueError('`groups` must be a list of lists of atoms in each group.')
        nuc = len(groups)
        range_uc = lambda uc: groups[uc]
    else:
        if nuc is not None:
            if not natoms%nuc==0.:
                raise ValueError('Number of unit cells are not a dividend of atoms.')
            nuc_atoms = natoms//nuc;
        else:
            nuc = 1
            nuc_atoms = natoms
        range_uc = lambda uc: range(uc*nuc_atoms, (uc+1)*nuc_atoms)
    
    indices = []
    for uc in range(nuc):
        indices.append(list(
            list(filter(
                lambda a: symbols[a]==elem, range_uc(uc)
                )) for elem in elements))
    
    perm_indices = []
    for uc in indices:
        for elem in uc:
            perm_indices.extend(
                basis.get_indices(elem).reshape(len(elem),-1).T.flatten())
    
    U = np.eye(basis.nao).take(perm_indices, axis=1)
    return U, indices


def get_ll_mm_rr(H, S, nbf_l, nbf_m, nbf_r):
    rng_l = list(range(nbf_l))
    rng_m = list(range(nbf_l,nbf_l+nbf_m))
    rng_r = list(range(nbf_l+nbf_m,nbf_l+nbf_m+nbf_r))
    s_lm = S.take(rng_l,0).take(rng_m,1)
    h_lm = H.take(rng_l,0).take(rng_m,1)
    s_rm = S.take(rng_r,0).take(rng_m,1)
    h_rm = H.take(rng_r,0).take(rng_m,1)
    h_mm = get_subspace(H, rng_m)
    s_mm = get_subspace(S, rng_m)
    return (h_lm, s_lm), (h_mm, s_mm), (h_rm, s_rm)


def get_lead_orthogonal(greenfunction):
    s_ll = greenfunction.selfenergies[0].s_ii
    s_rr = greenfunction.selfenergies[1].s_ii
    h_ll = greenfunction.selfenergies[0].h_ii
    h_rr = greenfunction.selfenergies[1].h_ii
    s_lm = greenfunction.selfenergies[0].s_im
    s_rm = greenfunction.selfenergies[1].s_im
    h_lm = greenfunction.selfenergies[0].h_im
    h_rm = greenfunction.selfenergies[1].h_im
    h_mm = greenfunction.H
    s_mm = greenfunction.S
    nbf_l = len(h_ll)
    nbf_r = len(h_rr)
    nbf_m = len(h_mm)
    s_ml = dagger(s_lm)
    s_mr = dagger(s_rm)
    h_ml = dagger(h_lm)
    h_mr = dagger(h_rm)
    U_lm = - la.inv(s_ll).dot(s_lm)
    U_rm = - la.inv(s_rr).dot(s_rm)
    U = np.array(block([[np.eye(nbf_l),U_lm,0],[0,np.eye(nbf_m),0],[0,U_rm,np.eye(nbf_r)]]))
    S = np.array(block([[s_ll,s_lm,0],[s_ml,s_mm,s_mr],[0,s_rm,s_rr]]))
    H = np.array(block([[h_ll,h_lm,0],[h_ml,h_mm,h_mr],[0,h_rm,h_rr]]))
    S = rotate_matrix(S,U)
    H = rotate_matrix(H,U)
    return get_ll_mm_rr(H, S, nbf_l, nbf_m, nbf_r)